@@ -11,16 +11,16 @@
 
 #include "mailer.h"
 #include <ctype.h>
-#include <signal.h>
 #include <fcntl.h>
 #include <sys/file.h>			/* O_RDONLY for run_praliases() */
 #include <pwd.h>			/* for run_homedir() */
 #include <grp.h>			/* for run_grpmems() */
-#include <mail.h>
-#include "interpret.h"
-#include "io.h"
 #include <errno.h>
 #include NDIR_H
+#include "zmsignal.h"
+#include "mail.h"
+#include "interpret.h"
+#include "io.h"
 
 #include "prototypes.h"
 
@@ -101,12 +101,12 @@ struct shCmd fnctns[] = {
 {	"rfc822",	run_rfc822,		0,		0	},
 {	"groupmembers",	run_grpmems,		0,		0	},
 {	"printaliases",	run_praliases,		0,		0	},
-{	"listaddresses",run_listaddrs,		0,		0	},
+{	"listaddresses",run_listaddrs,		0,		SH_ARGV	},
 {	"listexpand",	0,			run_listexpand,	SH_ARGV	},
 {	"homedirectory",run_homedir,		0,		0	},
 {	"rfc822date",	run_822date,		0,		0	},
-{	"filepriv",	run_filepriv,		0,		0	},
-{	"runas",	run_runas,		0,		0	},
+{	"filepriv",	run_filepriv,		0,		SH_ARGV	},
+{	"runas",	run_runas,		0,		SH_ARGV	},
 {	"gensym",	run_gensym,		0,		0	},
 {	"uid2login",	run_uid2login,		0,		0	},
 {	"login2uid",	run_login2uid,		0,		0	},
@@ -470,7 +470,7 @@ int sigarg;
 {
 	gothup = 1;
 	/* fprintf(stderr,"HUP\n"); */
-	signal(SIGHUP, sig_hup);
+	SIGNAL_HANDLE(SIGHUP, sig_hup);
 }
 
 void
@@ -644,6 +644,7 @@ rd_stability(dirp,dirs)
 	int namelen;
 	struct NDIR_TYPE *dp;
 	struct stat statbuf;
+	char pathbuf[512]; /* Enough ? */
 
 	deindex = 0;
 	nbindex = 0;
@@ -660,7 +661,8 @@ rd_stability(dirp,dirs)
 			continue;
 
 		/* See that the file is a regular file! */
-		if (lstat(dp->d_name,&statbuf) != 0) continue; /* ??? */
+		sprintf(pathbuf,"%s%s",dirs,dp->d_name);
+		if (lstat(pathbuf,&statbuf) != 0) continue; /* ??? */
 		if (!S_ISREG(statbuf.st_mode)) continue; /* Hmm..  */
 
 		if (deindex >= desize) {
@@ -688,8 +690,8 @@ rd_stability(dirp,dirs)
 		 */
 		dearray[deindex].f_stat =
 			(struct stat *)emalloc(sizeof (struct stat));
-		if (stat(dp->d_name, dearray[deindex].f_stat) < 0) {
-			perror(dp->d_name);
+		if (stat(pathbuf, dearray[deindex].f_stat) < 0) {
+			perror(pathbuf);
 			free((char *)dearray[deindex].f_stat);
 			continue;
 		}
@@ -731,6 +733,7 @@ rd_instability(dirp, dirs)
 	struct NDIR_TYPE *dp;
 	int did_cnt = 0;
 	struct stat statbuf;
+	char pathbuf[512];
 
 	while ((dp = readdir(dirp)) != NULL) {
 		if (mustexit)
@@ -745,11 +748,11 @@ rd_instability(dirp, dirs)
 			continue;
 
 		/* See that the file is a regular file! */
-		if (stat(dp->d_name,&statbuf) != 0) continue; /* ??? */
+		sprintf(pathbuf,"%s%s",dirs,dp->d_name);
+		if (stat(pathbuf,&statbuf) != 0) continue; /* ??? */
 		if (!S_ISREG(statbuf.st_mode)) continue; /* Hmm..  */
 
-		if (dp->d_name[0] != '.')
-			did_cnt += rd_doit(dp->d_name, dirs);
+		did_cnt += rd_doit(dp->d_name, dirs);
 	}
 	return did_cnt;
 }
@@ -792,11 +795,11 @@ run_daemon(argc, argv)
 	char pathbuf[256];
 	memtypes oval = stickymem;
 
-	signal(SIGTERM, sig_exit);	/* mustexit = 1 */
+	SIGNAL_HANDLE(SIGTERM, sig_exit);	/* mustexit = 1 */
 	for (i=0; i<ROUTERDIR_CNT; ++i) {
 		dirp[i] = NULL; dirs[i] = NULL;
 	}
-	dirp[0] = opendir(".");	/* assert dirp != NULL ... */
+	/* dirp[0] = opendir(".");	/* assert dirp != NULL ... */
 #ifdef BSD
 	dirp[0]->dd_size = 0;	/* stupid Berkeley bug workaround */
 #endif
@@ -809,8 +812,6 @@ run_daemon(argc, argv)
 			s = strchr(rd,':');
 			if (s)  *s = 0;
 			sprintf(pathbuf,"../%s",rd);
-			if ((dirp[i] = opendir(pathbuf)) == NULL)
-			  break;
 			strcat(pathbuf,"/");
 #ifdef	BSD
 			dirp[i]->dd_size = 0;
@@ -829,8 +830,24 @@ run_daemon(argc, argv)
 		rd_initstability();
 	stickymem = oval;
 	did_cnt = 0;
-	i = 0;
+	i = -1;
 	for (;;) {
+		++i;	/* Increment it */
+		/* The last of the alternate dirs ?  Reset.. */
+		if (i >= ROUTERDIR_CNT || dirs[i] == NULL) {
+			i = 0;
+
+			canexit = 1;
+		/*
+		 * If a shell signal interrupts us here, there is
+		 * potential for problems knowing which file descriptors
+		 * are free.  One could add a setfreefd() to the trap
+		 * routine, in that case.
+		 */
+			sleep(sweepintvl);
+			canexit = 0;
+		}
+
 		/*
 		 * We would like to do a seekdir()/rewinddir()
 		 * instead of opendir()/closedir()  inside this
@@ -844,6 +861,7 @@ run_daemon(argc, argv)
 		 */
 		/* rewinddir(dirp[i]); /* some system w/o this ? */
 
+#if 0
 #ifdef	BUGGY_CLOSEDIR
 		/*
 		 * Major serious bug time here;  some closedir()'s
@@ -853,6 +871,7 @@ run_daemon(argc, argv)
 		close(dirp[i]->dd_fd);
 #endif
 		closedir(dirp[i]);
+#endif
 		dirp[i] = opendir(dirs[i][0] == 0 ? "." : dirs[i]);
 
 		if (stability)
@@ -867,28 +886,7 @@ run_daemon(argc, argv)
 			else
 				rd_initstability();
 		}
-
-		if (mustexit)
-			break;
-
-		/* Alter router directory.  If processed directory
-		   had any job, reset the  index  to the begin.   */
-		++i;
-		if (did_cnt > 0) i = 0;
-		if (!dirp[i]) i = 0;
-		else continue;
-
-		canexit = 1;
-		/*
-		 * If a shell signal interrupts us here, there is
-		 * potential for problems knowing which file descriptors
-		 * are free.  One could add a setfreefd() to the trap
-		 * routine, in that case.
-		 */
-		sleep(sweepintvl);
-		canexit = 0;
-	}
-	for (i=0; dirp[i] != NULL && i < ROUTERDIR_CNT; ++i) {
+#if 1
 #ifdef	BUGGY_CLOSEDIR
 		/*
 		 * Major serious bug time here;  some closedir()'s
@@ -898,6 +896,28 @@ run_daemon(argc, argv)
 		close(dirp[i]->dd_fd);
 #endif
 		closedir(dirp[i]);
+#endif
+
+		if (mustexit)
+			break;
+
+		/* Alter router directory.  If processed directory
+		   had any job, reset the  index  to the begin.   */
+		if (did_cnt > 0)
+			i = -1;
+	}
+	for (i=0; dirp[i] != NULL && i < ROUTERDIR_CNT; ++i) {
+#if 0
+#ifdef	BUGGY_CLOSEDIR
+		/*
+		 * Major serious bug time here;  some closedir()'s
+		 * free dirp before referring to dirp->dd_fd. GRRR.
+		 * XX: remove this when bug is eradicated from System V's.
+		 */
+		close(dirp[i]->dd_fd);
+#endif
+		closedir(dirp[i]);
+#endif
 		free(dirs[i]);
 	}
 	return 0;
@@ -1368,8 +1388,9 @@ run_listexpand(avl, il)
 		siofds[fileno(mfp)] = NULL;
 		fprintf(mfp, "channel error\n");
 		fprintf(mfp, "to %s\n", erroraddress);
-		fprintf(mfp, "From: Error Channel <>\n");
+		fprintf(mfp, "From: Error Channel <MAILER-DAEMON>\n");
 		fprintf(mfp, "To: %s\n", erroraddress);
+		fprintf(mfp, "Cc: The Post Office <postmaster>\n");
 		fprintf(mfp, "Subject: Error in %s\n", comment);
 		fprintf(mfp, "Precedence: junk\n\n");
 		hdr_errprint(e, &hs, mfp, comment);
@@ -1668,9 +1689,11 @@ run_listaddrs(argc, argv)
 		siofds[fileno(mfp)] = NULL;
 		fprintf(mfp, "channel error\n");
 		fprintf(mfp, "to %s\n", erroraddress);
-		fprintf(mfp, "From: Error Channel <>\n");
+		fprintf(mfp, "From: Error Channel <MAILER-DAEMON>\n");
 		fprintf(mfp, "To: %s\n", erroraddress);
-		fprintf(mfp, "Subject: Error in %s\n\n", comment);
+		fprintf(mfp, "Cc: The Post Office <postmaster>\n");
+		fprintf(mfp, "Subject: Error in %s\n", comment);
+		fprintf(mfp, "Precedence: junk\n\n");
 		hdr_errprint(e, &hs, mfp, comment);
 		siofds[fileno(mfp)] = osiop;
 		mail_close(mfp);
