@@ -186,11 +186,10 @@ search_res(sip)
 	else
 	  _res.options &= ~RES_DEBUG;
 	/* [Edwin Allum]
-	 * Turn down the resolver's timeout and retry counts.
-	 * The defaults (6 and 4) will cause the router to fall hopelessly
-	 * behind in the face of broken DNS data.
+	 * Turn down the resolver's retry count.
+	 * The default (4) will cause the router to fall
+	 * hopelessly behind in the face of broken DNS data.
 	 */
-	_res.retrans = 3;
 	_res.retry   = 2;
 	if (sip->subtype == NULL || *(sip->subtype) == '\0') {
 	  fprintf(stderr, "search_res: missing subtype to BIND query!\n");
@@ -399,7 +398,7 @@ getmxrr(host, localhost, ttlp, depth)
 		if (n < 0)
 			break;
 		cp += n;
-		mx[nmx].host = (msgdata *)strsave(hbuf);
+		mx[nmx].host = (msgdata *)strsave((char*)hbuf);
 		if (localhost != NULL && CISTREQ(hbuf, localhost))
 			maxpref = mx[nmx].pref;
 		++nmx;
