@@ -14,6 +14,7 @@
 #ifdef	USE_MMAP
 #include <sys/mman.h>
 #endif
+#include <errno.h>
 
 extern int cistrcmp();
 extern int deferit;
@@ -68,6 +69,15 @@ reopen:
 		fm->offsets = NULL;
 		fm->membuf = (void*)mmap(NULL, fst.st_size,
 					 PROT_READ, MAP_SHARED, fileno(fp), 0);
+		if ((long)fm->membuf == -1) {
+			fprintf(stderr,
+				"search_seq: cannot mmap() (r/o) %d bytes of file \"%s\" into memory, errno=%d",
+				fst.st_size, sip->file, errno);
+			free(fm);
+			++deferit;
+			v_set(DEFER, DEFER_IO_ERROR);
+			return NULL;
+		}
 #else
 		fm->size   = 0;
 		fm->mtime  = 0;
