@@ -12,7 +12,7 @@
 #include <fcntl.h>
 #include <sys/file.h>
 #include <errno.h>
-#include <mail.h>
+#include "mail.h"
 #ifdef	USE_SYSLOG
 #include <syslog.h>
 #endif	/* USE_SYSLOG */
