@@ -211,8 +211,8 @@ copySender(e)
 	best = NULL;
 	minval = 10000;
 	for (h = e->e_headers; h != NULL; h = h->h_next) {
-		if (h->h_stamp == BadHeader) continue;
 		if (h->h_descriptor->user_type != Sender
+		    || h->h_stamp == BadHeader
 		    || (e->e_resent != 0 && h->h_descriptor->class != Resent))
 			continue;
 		cp = h->h_descriptor->hdr_name + e->e_resent;
