@@ -14,7 +14,7 @@
 #include <ctype.h>
 #include <fcntl.h>
 #include <sys/file.h>
-#include <mail.h>
+#include "mail.h"
 #ifdef	USE_SYSLOG
 #include <syslog.h>
 #endif	/* USE_SYSLOG */
