@@ -8,7 +8,7 @@
 #include <ctype.h>
 #include <sys/param.h>
 #include <fcntl.h>
-#include <mail.h>
+#include "mail.h"
 #include "scheduler.h"
 #ifdef	USE_SYSLOG
 #include <syslog.h>
@@ -38,8 +38,8 @@ struct not {
 	char *dsn;
 };
 
-static void
-notaryreport(errfp,notary)
+static void	/* There is notaryreport() on transporters also.. */
+scnotaryreport(errfp,notary)
 	FILE *errfp;
 	struct not *notary;
 {
@@ -50,6 +50,9 @@ notaryreport(errfp,notary)
 	}
 
 	rcpt     = notary->not;
+	if (*rcpt == 0)
+	  return;	/* XX: ?? eh, well.. we should not be called.. */
+
 	action   = strchr(rcpt,'\001');
 	*action++ = 0;
 	status   = strchr(action,'\001');
@@ -195,13 +198,13 @@ reporterrs(cfpi)
 
 	  cp = cfp->contents + *lp + 2;
 	  byteidx = atoi(cp);
-	  while (isascii(*cp) && isdigit(*cp))
+	  while (*cp  && isascii(*cp) && isdigit(*cp))
 	    ++cp;
 	  headeridx = -1;
 	  if (*cp == ':') {
 	    ++cp;
 	    headeridx = atoi(cp);
-	    while (isascii(*cp) && (isdigit(*cp) || *cp == '-'))
+	    while (*cp && isascii(*cp) && (isdigit(*cp) || *cp == '-'))
 	      ++cp;
 	  }
 	  drptidx = -1;
@@ -218,14 +221,15 @@ reporterrs(cfpi)
 	    notary = strnsave(notary, cp-notary);
 	    ++cp;
 	  }
-	  if (notary) {
+	  if (notary != NULL && *notary != 0) {
 	    if (!notaries) {
 	      notaryspc = 8;
-	      notaries = emalloc(sizeof(struct not)*(notaryspc+1));
+	      notaries = (void*)emalloc(sizeof(struct not)*(notaryspc+1));
 	    }
 	    if (notarycnt >= notaryspc) {
 	      notaryspc += 8;
-	      notaries = erealloc(notaries,sizeof(struct not)*(notaryspc+1));
+	      notaries = (void*)erealloc((void*)notaries,
+					 sizeof(struct not)*(notaryspc+1));
 	    }
 	    notaries[notarycnt].not = notary;
 	    notaries[notarycnt].dsn = NULL;
@@ -273,9 +277,9 @@ reporterrs(cfpi)
 
 	/* Now scan 'em all again for IETF-NOTARY */
 	for (i = 0; i < notarycnt; ++i) {
-	  notaryreport(errfp, &notaries[i]);
+	  scnotaryreport(errfp, &notaries[i]);
 	}
-	free(notaries);
+	free((void*)notaries);
 
 	fprintf(errfp, "--%s\n", boundarystr);
 	fprintf(errfp, "Content-Type: message/rfc822\n\n");
