@@ -14,23 +14,11 @@
 #include <errno.h>
 #include "scheduler.h"
 #include <sys/stat.h>
-#include <signal.h>
 #include <syslog.h>
 #include <fcntl.h>
 #include <mail.h>
 #include NDIR_H
-
-#ifdef	SV_INTERRUPT
-#define	SIGNAL_SYSCALLINTERRUPT(X,Y)	\
-	{	struct sigvec sv; \
-		sigvec(X, (struct sigvec *)NULL, &sv); \
-		sv.sv_handler = Y; \
-		sv.sv_flags |= SV_INTERRUPT; \
-		sigvec(X, &sv, (struct sigvec *)NULL); \
-	}
-#else	/* !SV_INTERRUPT */
-#define	SIGNAL_SYSCALLINTERRUPT(X,Y)	signal(X,Y)
-#endif	/* SV_INTERRUPT */
+#include "zmsignal.h"
 
 #ifndef	MAXNAMLEN /* POSIX.1 ... */
 #define MAXNAMLEN NAME_MAX
@@ -216,14 +204,14 @@ main(argc, argv)
 		if (loginit() < 0)	/* do setlinebuf() there */
 			die(1, "log initialization failure");
 		/* close and reopen log files */
-		SIGNAL_SYSCALLINTERRUPT(SIGHUP, (SIGNAL_TYPE (*)())loginit);
+		SIGNAL_HANDLE(SIGHUP, (SIGNAL_TYPE (*)())loginit);
 	} else {
-		SIGNAL_SYSCALLINTERRUPT(SIGHUP, SIG_IGN); /* no surprises please */
+		SIGNAL_IGNORE(SIGHUP); /* no surprises please */
 		setvbuf(stdout, (char *)NULL, _IOLBF, 0);
 		setbuf(stderr,NULL); /* [mea] No buffering on stdout.. */
 	}
 #ifdef	SIGUSR1
-	SIGNAL_SYSCALLINTERRUPT(SIGUSR1, sig_readcf);
+	SIGNAL_HANDLE(SIGUSR1, sig_readcf);
 #endif	/* SIGUSR1 */
 	if (verbose || version) {
 		prversion("scheduler");
@@ -297,9 +285,9 @@ main(argc, argv)
 	}
 	mustexit = gotalarm = dumpq = rereadcf = 0;
 	canexit = 0;
-	SIGNAL_SYSCALLINTERRUPT(SIGALRM, sig_alarm);	/* process agenda */
-	SIGNAL_SYSCALLINTERRUPT(SIGPIPE, SIG_IGN);
-	SIGNAL_SYSCALLINTERRUPT(SIGUSR2, sig_iot);	/* dump queue info */
+	SIGNAL_IGNORE(SIGPIPE);
+	SIGNAL_HANDLE(SIGALRM, sig_alarm);	/* process agenda */
+	SIGNAL_HANDLE(SIGUSR2, sig_iot);	/* dump queue info */
 	cp = emalloc(strlen(TRANSPORTDIR)+4);
 	sprintf(cp, "../%s", TRANSPORTDIR);
 	queryipcinit();
@@ -314,7 +302,7 @@ main(argc, argv)
 	next_delaysweep += 1*60; /* One minute */
 
 	canexit = 1;
-	SIGNAL_SYSCALLINTERRUPT(SIGTERM, sig_exit);	/* split */
+	SIGNAL_HANDLE(SIGTERM, sig_exit);	/* split */
 	sweepretry = 1;
 	alarm(sweepinterval);
 #ifdef	MALLOC_TRACE
@@ -395,7 +383,7 @@ SIGNAL_TYPE
 sig_alarm()
 {
 	gotalarm = 1;
-	SIGNAL_SYSCALLINTERRUPT(SIGALRM, sig_alarm);
+	SIGNAL_HANDLE(SIGALRM, sig_alarm);
 }
 
 #ifdef SIGUSR2
@@ -403,7 +391,7 @@ SIGNAL_TYPE
 sig_iot()
 {
 	dumpq = 1;
-	SIGNAL_SYSCALLINTERRUPT(SIGUSR2, sig_iot);
+	SIGNAL_HANDLE(SIGUSR2, sig_iot);
 }
 #endif
 
@@ -412,7 +400,7 @@ SIGNAL_TYPE
 sig_readcf()
 {
 	rereadcf = 1;
-	SIGNAL_SYSCALLINTERRUPT(SIGUSR1, sig_readcf);
+	SIGNAL_HANDLE(SIGUSR1, sig_readcf);
 }
 #endif
 
