@@ -25,7 +25,10 @@
 
 #ifdef	USE_UNIONWAIT
 #include <sys/wait.h>
-#endif	/* USE_UNIONWAIT */
+#endif
+#ifdef	USE_WAITPID
+#include <sys/wait.h>
+#endif
 
 struct procinfo *cpids = NULL;
 
@@ -411,9 +414,9 @@ waitandclose(fd)
 	if (cpids[fd].pid > 0) {
 #ifdef	USE_UNIONWAIT
 	  while ((pid = wait((union wait *)0)) > 0)
-#else				/* !USE_UNIONWAIT */
+#else	/* !USE_UNIONWAIT */
 	  while ((pid = wait((int *)0)) > 0)
-#endif				/* USE_UNIONWAIT */
+#endif	/* USE_UNIONWAIT */
 	    {
 	      --numkids;
 	      if (verbose)
