@@ -32,36 +32,41 @@ char *Copyright2= "Copyright 1991-1994 Matti Aarnio";
 #include "hostenv.h"
 #include "malloc.h"
 #include <sys/types.h>
+#include <ctype.h>
+#include <sys/time.h>
 #include <sys/stat.h>
 #include <sys/file.h>
-#include <ctype.h>
-#include <errno.h>
-#include <signal.h>
-#include <varargs.h>
-#include <mail.h>
-#include <sys/time.h>
 #ifdef FCNTL_H
 # include FCNTL_H
 #endif
+#include "zmsignal.h"
+#include <errno.h>
+#include <varargs.h>
+#include "mail.h"
 
 #ifdef	USE_UNIONWAIT
-#include <sys/wait.h>
-#endif	/* USE_UNIONWAIT */
-#ifdef	WNOHANG
-#include <sys/resource.h>
-#endif	/* WNOHANG */
+# include <sys/wait.h>
+# ifdef	WNOHANG
+#  include <sys/resource.h>
+# endif
+#endif
+
+#ifdef USE_WAITPID
+# include <sys/wait.h>
+#endif
+
 #if	defined(USE_HOSTS) || defined(USE_INET)
-#include <netdb.h>
+# include <netdb.h>
 #endif	/* USE_HOSTS */
 #ifdef	USE_INET
-#include <sys/socket.h>
-#include <netinet/in.h>
+# include <sys/socket.h>
+# include <netinet/in.h>
 #endif	/* USE_INET */
 #ifdef	USE_SYSLOG
-#include <syslog.h>
+# include <syslog.h>
 #endif	/* USE_SYSLOG */
 #ifdef	USE_AUTH
-#include <authuser.h>
+# include <authuser.h>
 #endif	/* USE_AUTH */
 
 #ifndef	SIGCHLD
@@ -130,6 +135,7 @@ struct command {
 
 struct smtpconf {
 	char	*pattern;
+	int	maxloadavg;
 	char	*flags;
 	struct smtpconf *next;
 };
@@ -179,14 +185,19 @@ char	*with_protocol = WITH_SMTP;
 struct command *carp = NULL;
 FILE *mfp = NULL;
 struct smtpconf *cfinfo = NULL;
-long availspace = 0;	/* available diskspace/2 in bytes	*/
+long availspace = -1;	/* available diskspace/2 in bytes	*/
+long maxsize    = 0;
 long sizeoptval = 0;	/* "MAIL FROM:<xxx> SIZE=nnn" -value	*/
+long sizeoptsum = 0;
 char buf[SMTPLINESIZE];	/* limits size of SMTP commands...
 			   On the other hand, limit is asked
 			   to be only 1000 chars, not 8k.. */
 char helobuf[SMTPLINESIZE];
-int maxloadavg = 10;	/* Maximum load-average that is tolerated
-			   with smtp-server actively receiving.. */
+int maxloadavg = 999;	/* Maximum load-average that is tolerated
+			   with smtp-server actively receiving..
+			   Default value of 999 is high enough
+			   so that it will never block -- use
+			   "-L 10" to define lower limit (10) */
 
 
 
@@ -215,7 +226,8 @@ extern char *dottedquad();
 extern char *Version;
 extern char *VersionNumb;
 extern u_char *rfc822date();
-extern void mvdata(), type(), report(), debug_report();
+extern int mvdata();
+extern void type(), report(), debug_report();
 extern void help();
 extern time_t time();
 extern char *router();
@@ -240,7 +252,7 @@ extern char *strchr();
 #define	putc	fputc
 #endif	/* lint */
 
-char *msg_toohighload = "400 Sorry, system is too loaded for email reception at the moment\r\n"; /* XX: ??? */
+char *msg_toohighload = "421 Sorry, the system is too loaded for email reception at the moment\r\n"; /* XX: ??? */
 
 
 void
@@ -293,7 +305,7 @@ main(argc, argv)
 				       progname, myhostname, strerror(errno));
 		exit(1);
 	}
-	while ((c = getopt(argc, argv, "aBigl:np:L:P:R:s:Vv")) != EOF) {
+	while ((c = getopt(argc, argv, "aBigl:np:L:M:P:R:s:Vv")) != EOF) {
 		switch (c) {
 #ifdef	USE_AUTH
 		case 'a':
@@ -303,6 +315,11 @@ main(argc, argv)
 		case 'B':
 			with_protocol = WITH_BSMTP;
 			break;
+		case 'M':
+			maxsize = atoi(optarg);
+			if (maxsize < 0)
+			  maxsize = 0;
+			break;
 		case 'i':	/* interactive */
 			daemon = 0;
 			break;
@@ -365,11 +382,11 @@ main(argc, argv)
 	if (errflg || optind != argc) {
 #ifdef	USE_INET
 		fprintf(stderr,
-			"Usage: %s [-aBivgnV] [-s xx] [-L maxLoadAvg] [-R rtrprog] [-p port#] [-P postoffice] [-l logfile]\n",
+			"Usage: %s [-aBivgnV] [-s xx] [-L maxLoadAvg] [-M SMTPmaxsize] [-R rtrprog] [-p port#] [-P postoffice] [-l logfile]\n",
 			progname);
 #else	/* !USE_INET */
 		fprintf(stderr,
-			"Usage: %s [-aBivgnV] [-s xx] [-L maxLoadAvg] [-R rtrprog] [-P postoffice] [-l logfile]\n",
+			"Usage: %s [-aBivgnV] [-s xx] [-L maxLoadAvg] [-M SMTPmaxsize] [-R rtrprog] [-P postoffice] [-l logfile]\n",
 			progname);
 #endif	/* USE_INET */
 		exit(1);
@@ -498,13 +515,13 @@ main(argc, argv)
 				close(msgfd);
 			}
 #if 0
-		signal(SIGCHLD, SIG_IGN);
+		SIGNAL_HANDLE(SIGCHLD, SIG_IGN);
 #else
-		signal(SIGCHLD, reaper);
+		SIGNAL_HANDLE(SIGCHLD, reaper);
 #endif
-		signal(SIGALRM, timedout);
-		signal(SIGHUP,  SIG_IGN);
-		signal(SIGTERM, SIG_DFL);
+		SIGNAL_HANDLE(SIGALRM, timedout);
+		SIGNAL_HANDLE(SIGHUP,  SIG_IGN);
+		SIGNAL_HANDLE(SIGTERM, SIG_DFL);
 		while (1) {
 			int childpid;
 			raddrlen = sizeof raddr;
@@ -536,7 +553,7 @@ main(argc, argv)
 				  exit(1);
 				}
 
-				signal(SIGTERM, SIG_IGN);
+				SIGNAL_HANDLE(SIGTERM, SIG_IGN);
 				if (msgfd != 0)
 					dup2(msgfd, 0);
 				if (msgfd != 1)
@@ -674,13 +691,17 @@ reaper()
 	int status;
 #endif	/* USE_UNIONWAIT */
 
-#ifdef	WNOHANG
+#ifdef	USE_WAITPID
+	while (waitpid(-1, &status, WNOHANG) > 0)
+#else
+#ifdef	WNOHANG /* UNIONWAIT with WNOHANG .. */
 	while (wait3(&status, WNOHANG, (struct rusage *)NULL) > 0)
 #else	/* !WNOHANG */
 	while (wait(&status) > 0)
 #endif	/* WNOHANG */
+#endif
 		continue;
-	signal(SIGCHLD, reaper);
+	SIGNAL_HANDLE(SIGCHLD, reaper);
 }
 
 void
@@ -763,17 +784,14 @@ char *cp;
 	  type(carp->cmd==Hello2 ? -250: 250,
 	       " %s Hello %s", myhostname, helobuf);
 	if (carp->cmd == Hello2) {
-		/*  -- this is NOT useful, while we do not report
-		       available disk space on the ESMTP initial greeting..
-		   if (mfp == NULL &&
-		   (mfp = mail_open(MSG_RFC822)) == NULL) {
-		   type(452, (char *)NULL);
-		   return;
-		   }
-		  */
 
 		/* ESMTP -- RFC 1651,1652,1653,1428 thingies */
-		type(-250,"SIZE");
+		char sizebuf[20];
+
+		sprintf(sizebuf,"SIZE %d",maxsize); /* 0: No fixed max size
+						          in force, else:
+							  The FIXED maximum */
+		type(-250,sizebuf);
 		type(-250,"8BITMIME");
 		type(-250,"EXPN");
 		type(-250,"VRFY");
@@ -806,6 +824,7 @@ int insecure;
 		type(452, (char *)NULL);
 		return;
 	}
+
 	availspace = fd_statfs(fileno(mfp));
 	if (availspace < 0)
 	  availspace = 2000000000; /* Over 2G ? */
@@ -864,6 +883,7 @@ int insecure;
 
 	/* BODY=8BITMIME SIZE=nnnn ENVID=xxxxxx */
 	sizeoptval = -1;
+	sizeoptsum = -1;
 	drpt_envid = NULL;
 	rc = 0;
 	while (*s) {
@@ -977,6 +997,8 @@ int insecure;
 	if (ferror(mfp)) {
 		type(452, (char *)NULL);
 		clearerr(mfp);
+	} else if (sizeoptval > maxsize && maxsize > 0) {
+		type(552, "The message size exceeds fixed maximum message size");
 	} else if (sizeoptval > availspace) {
 		type(452, "Try again later, insufficient storage available at the moment");
 	} else if (s)
@@ -1143,11 +1165,11 @@ char *cp;
 	fprintf(mfp, "to <%s>\n", cp);
 	if (sizeoptval < 0)
 		sizeoptval = 0;
-	availspace -= sizeoptval;
+	sizeoptsum += sizeoptval;
 	if (ferror(mfp)) {
 		type(452, (char *)NULL);
 		clearerr(mfp);
-	} else if (availspace < 0) {
+	} else if (sizeoptsum > availspace) {
 		type(452, "insufficient storage space, try again later");
 	} else if (s) {
 		type(atoi(s), s+4, "Ok", "Ok");
@@ -1169,7 +1191,7 @@ smtpserver(insecure)
 {
 	char *cp, *s;
 	time_t now;
-	long tell;
+	long tell, filsiz;
 	int inum;
 	struct stat stbuf;
 	struct sockaddr_in localsock;
@@ -1365,7 +1387,7 @@ smtpserver(insecure)
 			}
 			type(354, (char *)NULL);
 			alarm(60*60*24);	/* one day should be enough! */
-			mvdata(mfp);
+			filsiz = mvdata(mfp);
 			alarm(0);	/* cancel alarm */
 			tell = ftell(mfp);
 			if (fstat(fileno(mfp), &stbuf) < 0)
@@ -1383,6 +1405,10 @@ smtpserver(insecure)
 				clearerr(mfp);
 				mail_abort(mfp);
 				reporterr(tell, "message file error");
+			} else if (maxsize > 0 && filsiz > maxsize) {
+				mail_abort(mfp);
+				type(552, "Message size exceeded fixed maximum size for acceptable email");
+
 			} else if (mail_close(mfp) == EOF) {
 				type(452, (char *)NULL);
 				reporterr(tell, "message file close failed");
@@ -1768,11 +1794,11 @@ callr()
 		return -1;
 
 #if 0
-	signal(SIGCHLD, SIG_IGN);
+	SIGNAL_HANDLE(SIGCHLD, SIG_DFL); /* ?? SIG_DFL  on SVR4 ? */
 #else
-	signal(SIGCHLD, reaper);
+	SIGNAL_HANDLE(SIGCHLD, reaper);
 #endif
-	signal(SIGPIPE, SIG_IGN);
+	SIGNAL_HANDLE(SIGPIPE, SIG_IGN);
 
 	if (routerprog == NULL) {
 		if ((buf = getzenv("MAILBIN")) == NULL) {
@@ -2067,7 +2093,33 @@ int	states[] = {
  * that corresponds to the input character.
  */
 
-int	indexnum[256+1] = { 4 };
+static int	indexnum[256+1] = {
+#if 0
+	idxnum['\r'] = 1;
+	idxnum['\n'] = 2;
+	idxnum['.'] = 3;
+#if EOF == -1
+	idxnum[EOF] = 4;
+#endif
+#endif
+	4, /* EOF */
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 1, 0, 0, /* ...'\n'..'\r'.. */
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, /* ... '.' .. */
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
+};
 
 /*
  * Copy bytes from stdin to out, obeying sensible SMTP DATA input heuristics.
@@ -2076,42 +2128,119 @@ int	indexnum[256+1] = { 4 };
  * This version goes at better than 100kB/cpu-sec on a Sun 3/180.
  */
 
-void
+int /* count of bytes */
 mvdata(out)
 	register FILE *out;
 {
 	register FILE *input;
-	register int c, state, *idxnum, *sts, endstate;
+	register int c, state, *idxnum, *sts, endstate, cnt;
 
+#ifdef NO_INCOMING_HEADER_PROCESSING
 	idxnum = indexnum+1;
-	idxnum['\r'] = 1;
-	idxnum['\n'] = 2;
-	idxnum['.'] = 3;
-#if EOF == -1
-	idxnum[EOF] = 4;
-#endif
+
 	state = 10;
 	endstate = X_;
 	input = stdin;
 	sts = states;
+	cnt = 0;
+#else
+	char linebuf[2000], *s, *eol;
+	int col;
+
+	idxnum = indexnum+1;
+
+	state = 10;
+	endstate = X_;
+	input = stdin;
+	sts = states;
+	cnt = 0;
+	col = 0;
+
+	/* ================ Input the email headers ================ */
+	/*	     and analyze them a bit (Precedence:)	     */
+	mail_priority = _MAILPRIO_NORMAL;
 	for (;;) {
 		c = getc(input);
+		/* An EOF in here is an error! */
 #if EOF != -1
-		if (c == EOF)		/* a little slower... */
-			return;
+		if (c == EOF) return cnt;
 #endif
 		if ((state = sts[state+idxnum[c]]) & ~0xff) {
 			if (state & O_) {
 				if (state == endstate)
-					return;
+					return cnt;
 				state = (char)state;
 				continue;
 			}
-			putc((state>>8), out);
+			if (!ferror(out))
+				putc((state>>8), out);
+			if (col < (sizeof(linebuf)-1))
+				linebuf[col++] = (state >> 8);
 			state = (char)state;
 		}
-		putc(c, out);
+		if (!ferror(out))
+			putc(c, out);
+		if (col < (sizeof(linebuf)-1))
+			linebuf[col++] = c;
+		/* LF, or something else ?  Here the else.. */
+		if (c != '\n')
+			continue;
+
+		/* See what this header is about -- or if the body starts.. */
+		if (col > sizeof(linebuf)-1) col = sizeof(linebuf)-1;
+		eol = linebuf + col;
+		*eol = 0;
+		for (s = linebuf; s < eol; ++s)
+			if (*s != ' ' && *s != '\t' &&
+			    *s != '\r' && *s != '\n')
+				break;
+		if (s == eol) { /* All-blank line */
+			break;  /* Into the body processing */
+		}
+		col = 0;
+		if (*linebuf == ' ' || *linebuf == '\t')
+			continue;	/* continuation line.. */
+
+		/* ================ PROCESS THE HEADERS! ================ */
+		if (strncasecmp(linebuf,"Precedence:",11)==0) {
+			s = linebuf+11;
+			while (*s == ' ' || *s == '\t') ++s;
+			if ((eol - s) < 4) continue; /* Hmm.. */
+			if (strncasecmp("junk",s,4)==0) {
+				mail_priority = _MAILPRIO_JUNK;
+			} else if (strncasecmp("bulk",s,5)==0) {
+				mail_priority = _MAILPRIO_BULK;
+			}
+		}
 	}
+	/* if(logfp)
+	   fprintf(logfp,"%d#\t(mail_priority=%d)\n",pid,mail_priority);
+	 */
+#endif
+
+	/* ================ Normal email BODY input.. ================ */
+	for (;;) {
+		c = getc(input);
+#if EOF != -1
+		if (c == EOF)		/* a little slower... */
+			return cnt;
+#endif
+		++cnt;
+		if ((state = sts[state+idxnum[c]]) & ~0xff) {
+			if (state & O_) {
+				if (state == endstate)
+					return cnt;
+				state = (char)state;
+				continue;
+			}
+			if (!ferror(out))
+			  putc((state>>8), out);
+			state = (char)state;
+		}
+		if (!ferror(out))
+		  putc(c, out);
+	}
+	return cnt;
 }
 
 #ifdef	USE_AUTH
@@ -2158,15 +2287,21 @@ readcffile(name)
 		SKIPWHILE(!isspace,cp);
 		c = *cp;
 		*cp = '\0';
-		scf.pattern = emalloc(strlen(buf)+1);
-		strcpy(scf.pattern, buf);
+		scf.pattern = strdup(buf);
+		scf.maxloadavg = 999;
 		if (c != '\0') {
-			++cp;
-			SKIPWHILE(isspace,cp);
-			scf.flags = emalloc(strlen(cp)+1);
-			strcpy(scf.flags, cp);
-			if ((cp = strchr(scf.flags, '\n')) != NULL)
-				*cp = '\0';
+		  ++cp;
+		  SKIPWHILE(isspace,cp);
+		  if (*cp && isascii(*cp) && isdigit(*cp)) {
+		    /* Sanity-check -- 2 is VERY LOW */
+		    if ((scf.maxloadavg = atoi(cp)) < 2)
+		      scf.maxloadavg = 2;
+		    SKIPWHILE(isdigit,cp);
+		    SKIPWHILE(isspace,cp);
+		  }
+		  scf.flags = strdup(cp);
+		  if ((cp = strchr(scf.flags, '\n')) != NULL)
+		    *cp = '\0';
 		}
 		for (cp = scf.pattern; *cp != '\0'; ++cp)
 			if (isascii(*cp) && isalpha(*cp) && isupper(*cp))
@@ -2191,7 +2326,11 @@ findcf(h)
 	struct smtpconf *scfp;
 	register char *cp, *s;
 
+#ifndef	USE_ALLOCA
 	cp = emalloc(strlen(h)+1);
+#else
+	cp = alloca(strlen(h)+1);
+#endif
 	for (s = cp; *h != '\0'; ++h) {
 		if (isascii(*h) && isalpha(*h) && isupper(*h))
 			*s++ = tolower(*h);
@@ -2201,10 +2340,14 @@ findcf(h)
 	*s = '\0';
 	for (scfp = cfhead; scfp != NULL; scfp = scfp->next) {
 		if (strmatch(scfp->pattern, cp)) {
+#ifndef USE_ALLOCA
 			free(cp);
+#endif
 			return scfp;
 		}
 	}
+#ifndef USE_ALLOCA
 	free(cp);
+#endif
 	return NULL;
 }
