@@ -7,10 +7,10 @@
 #include "hostenv.h"
 #include <ctype.h>
 #include <pwd.h>
-#include <signal.h>
 #include <sysexits.h>
-/*#include <string.h>*/
 #include <sys/param.h>
+#include <sys/stat.h>
+#include "zmsignal.h"
 #include "malloc.h"
 #include NDIR_H
 
@@ -81,7 +81,7 @@ typedef char msgdata;
 #endif
 
 
-#include <mail.h>
+#include "mail.h"
 #include "splay.h"
 
 #ifdef	USE_UNIONWAIT
@@ -138,18 +138,23 @@ main(argc, argv)
 	char *channel, file[MAXPATHLEN+1];
 	int errflg, c, i;
 	struct ctldesc *dp;
+	SIGNAL_TYPE (*oldsig)();
 
 	pid = getpid();
 	cmdline = &argv[0][0];
 	eocmdline = cmdline;
 	for (i = 0; i < argc; ++i)
 	  eocmdline += strlen(argv[i]) + 1;
-	if (signal(SIGINT, SIG_IGN) != SIG_IGN)
-	  signal(SIGINT, wantout);
-	if (signal(SIGHUP, SIG_IGN) != SIG_IGN)
-	  signal(SIGHUP, wantout);
-	if (signal(SIGTERM, SIG_IGN) != SIG_IGN)
-	  signal(SIGTERM, wantout);
+
+	SIGNAL_HANDLESAVE(SIGINT, SIG_IGN, oldsig);
+	if (oldsig != SIG_IGN)
+	  SIGNAL_HANDLE(SIGINT, wantout);
+	SIGNAL_HANDLESAVE(SIGTERM, SIG_IGN, oldsig);
+	if (oldsig != SIG_IGN)
+	  SIGNAL_HANDLE(SIGTERM, wantout);
+	SIGNAL_HANDLESAVE(SIGHUP, SIG_IGN, oldsig);
+	if (oldsig != SIG_IGN)
+	  SIGNAL_HANDLE(SIGHUP, wantout);
 
 	if ((progname = strrchr(argv[0], '/')) == NULL)
 	  progname = argv[0];
@@ -329,7 +334,7 @@ process(dp)
  * string is canonicalized and parsed hierarchically.
  */
 
-extern int hold_ns(), hold_timeout(), hold_io(), hold_script();
+extern int hold_ns(), hold_timeout(), hold_io(), hold_script(), hold_home();
 
 struct holds_info {
 	char	*type;
@@ -339,6 +344,7 @@ struct holds_info {
 	{	"timeout",	hold_timeout	},
 	{	"io",		hold_io		},
 	{	"script",	hold_script	},
+	{	"home",		hold_home	},
 	{	NULL,		NULL		},
 };
 
@@ -609,9 +615,9 @@ hold_script(command)
 	    close(in[0]);
 	  if (in[1] != 1 && in[1] != 2)
 	    close(in[1]);
-	  signal(SIGINT, SIG_IGN);
-	  signal(SIGHUP, SIG_IGN);
-	  signal(SIGTERM, SIG_DFL);
+	  SIGNAL_IGNORE(SIGINT);
+	  SIGNAL_IGNORE(SIGHUP);
+	  SIGNAL_HANDLE(SIGTERM, SIG_DFL);
 	  /*
 	   * Note that argv[0] is set to the command we are running.
 	   * That way, we should get some better error messages, at
@@ -672,6 +678,25 @@ hold_script(command)
 	return 1;
 }
 
+
+/* The transient condition is the user's home directory is not available.
+   So if the user doesn't exist any more, or if his home directory is
+   null, then we "succeed", to try to redeliver the mail. */
+
+int
+hold_home(user)
+char *user;
+{
+	struct passwd *pw;
+	struct stat st;
+
+	if ((pw = getpwnam(user)) == NULL) return 1;
+	if (pw->pw_dir == NULL || pw->pw_dir[0] == '\0') return 1;
+	if (stat(pw->pw_dir, &st) == 0 &&
+	    S_ISDIR(st.st_mode)) return 1;
+	return 0;
+}
+
 #ifdef	BIND
 
 typedef union {
