@@ -18,7 +18,6 @@
 #include <ctype.h>
 #include <errno.h>
 #include <pwd.h>
-#include <signal.h>
 #include <sysexits.h>
 #include <varargs.h>
 #include <sys/param.h>
@@ -26,7 +25,8 @@
 #include <fcntl.h>
 #include <sys/file.h>
 #include <sys/stat.h>
-#include <mail.h>
+#include "mail.h"
+#include "zmsignal.h"
 #include "ta.h"
 #include "splay.h"
 #include "malloc.h"
@@ -244,8 +244,8 @@ extern void program __((struct ctldesc *dp, struct rcpt *rp, char *timestring));
 extern FILE *putmail __((struct ctldesc *dp, struct rcpt *rp, int fdmail, char *timestring, char *file));
 extern int appendlet __((struct ctldesc *dp, struct rcpt *rp, FILE *fp, char *file, int ismime));
 extern char **environ;
-extern int writebuf __((FILE *fp, char *buf, int len));
-extern int writemimebuf __((FILE *fp, char *buf, int len));
+extern int writebuf __((FILE *fp, char *buf, int len, int *lastchp));
+extern int writemimeline __((FILE *fp, char *buf, int len, int *lastchp));
 
 extern int  qp_to_8bit   __((struct rcpt *rp));
 extern int  qptext_check __((struct rcpt *rp));
@@ -283,14 +283,19 @@ main(argc, argv)
 #if	defined(BIFF) || defined(RBIFF)
 	struct biffer *nbp;
 #endif
+	SIGNAL_TYPE (*oldsig)();
 
-	if (signal(SIGINT, SIG_IGN) != SIG_IGN)
-	  signal(SIGINT, wantout);
-	if (signal(SIGHUP, SIG_IGN) != SIG_IGN)
-	  signal(SIGHUP, wantout);
-	if (signal(SIGTERM, SIG_IGN) != SIG_IGN)
-	  signal(SIGTERM, wantout);
-	signal(SIGPIPE, SIG_IGN);
+	SIGNAL_HANDLESAVE(SIGINT, SIG_IGN, oldsig);
+	if (oldsig != SIG_IGN)
+	  SIGNAL_HANDLE(SIGINT, wantout);
+	SIGNAL_HANDLESAVE(SIGTERM, SIG_IGN, oldsig);
+	if (oldsig != SIG_IGN)
+	  SIGNAL_HANDLE(SIGTERM, wantout);
+	SIGNAL_HANDLESAVE(SIGHUP, SIG_IGN, oldsig);
+	if (oldsig != SIG_IGN)
+	  SIGNAL_HANDLE(SIGHUP, wantout);
+
+	SIGNAL_IGNORE(SIGPIPE);
 
 	if ((progname = strrchr(argv[0], '/')) == NULL)
 	  progname = argv[0];
@@ -524,9 +529,11 @@ deliver(dp, rp, timestring)
 	register char **maild;
 	int fdmail, uid, messagefd;
 	FILE *fp;
+	int hasdir;
 	struct stat st, s2;
 	char *file = NULL, *cp, *at;
 	int havedotlock = 0, havemaillock = 0;
+	int ismbox = 0;
 #if	defined(BIFF) || defined(RBIFF)
 	struct biffer *nbp = NULL;
 #ifdef	RBIFF
@@ -589,6 +596,7 @@ deliver(dp, rp, timestring)
 	default:		/* local user */
 	  /* Zap the possible '@' for a moment -- and restore later
 	     [mea@utu.fi] */
+	  ismbox = 1;
 	  if (at) *at = 0;
 	  if ((pw = getpwnam(rp->addr->user)) == NULL) {
 	    if (isascii(*(rp->addr->user))
@@ -606,14 +614,24 @@ deliver(dp, rp, timestring)
 	      return;
 	    }
 	  }
+	  hasdir = 0;
 	  for (maild = &maildirs[0]; *maild != 0; maild++) {
+	    if (stat(*maild,&st) < 0 ||
+		!S_ISDIR(st.st_mode))
+	      /* Does not exist, or is not a directory */
+	      continue;
+	    hasdir = 1;
 	    if ((file = exists(*maild, rp)) != NULL)
 	      break;		/* found it */
 	    if (rp->status != EX_OK)
 	      return;
 	  }
-	  if (*maild == 0) {	/* didn't find it? */
-	    /* if (!creatembox(rp, &file, &st.st_uid, &st.st_gid, pw)) */
+
+	  if (hasdir && *maild == 0 &&
+	      !creatembox(rp, &file, &st.st_uid, &st.st_gid, pw))
+	    return; /* creatembox() sets status */
+
+	  if (!hasdir) {	/* No directory ?? */
 
 	    char *mailbox = getzenv("MAILBOX");
 	    notaryreport(rp->addr->user,"failed",
@@ -632,7 +650,7 @@ deliver(dp, rp, timestring)
 	    else
 	      syslog(LOG_ALERT, "ZMailer mailbox can't open local mail folder directory!  $MAILBOX='%s'\n", mailbox);
 	    closelog();
-	    return;	/* creatembox sets status */
+	    return;
 	  }
 #if	defined(BIFF) || defined(RBIFF)
 	  if (!dobiff && !dorbiff)
@@ -699,6 +717,18 @@ deliver(dp, rp, timestring)
 		     "too many links to file \"%s\"", file);
 	  return;
 	}
+	/* [Edwin Allum]
+	 * If we're delivering to a mailbox, and the mail spool directory
+	 * is group writable, set our gid to that of the directory.
+	 * This allows us to create lock files if need be, without
+	 * making the spool dir world-writable.
+	 */
+	if (ismbox && (cp = strrchr(file, '/')) != NULL) {
+	  *cp = '\0';
+	  if (stat(file, &s2) == 0 && (s2.st_mode&020))
+	    st.st_gid = s2.st_gid;
+	  *cp = '/';
+	}
 	if (!setupuidgid(rp, st.st_uid, st.st_gid)) {
 	  setrootuid(rp);
 	  return;			/* setupuidgid sets status */
@@ -763,7 +793,11 @@ deliver(dp, rp, timestring)
 	  else
 	    maillockuser = rp->addr->user;
 
-	  switch (maillock(maillockuser, 2)) {
+	  if (ismbox) {
+	    int i;
+	    char errbuf[30];
+
+	    switch ((i = maillock(maillockuser, 2))) {
 	    case L_SUCCESS:
 	      cp = NULL;
 	      break;
@@ -787,18 +821,20 @@ deliver(dp, rp, timestring)
 	      cp = "cannot set mandatory lock on temp lockfile";
 	      break;
 	    default:
-	      cp = "unknown maillock() error";
+	      sprintf(errbuf, "maillock() error %d", i);
+	      cp = errbuf;
 	      break;
+	    }
+	    if (cp != NULL) {
+	      DIAGNOSTIC3(rp, EX_UNAVAILABLE,
+			  "Error locking \"%s\": %s", file, cp);
+	      return;
+	    }
+	    havemaillock = 1;
 	  }
-          if (cp != NULL) {
-	    DIAGNOSTIC3(rp, EX_UNAVAILABLE,
-			"Error locking \"%s\": %s", file, cp);
-	    return;
-	  }
-	  havemaillock = 1;
 #endif	/* USE_MAILLOCK */
 #ifdef  USE_DOTLOCK
-	  if (*(rp->addr->user) != TO_FILE &&
+	  if (ismbox &&
 	      !(havedotlock = dotlock(file) == 0)) {
 	    char mbuf[256];
 	    sprintf(mbuf, "\"%s\", errno %d", file, errno);
@@ -939,6 +975,7 @@ putmail(dp, rp, fdmail, timestring, file)
 	char buf[2];
 	struct stat st;
 	char *fromuser;
+	int lastch;
 
 	if ((fp = fdopen(fdmail, "a+")) == NULL) {
 	  notaryreport(NULL,NULL,NULL);
@@ -948,6 +985,7 @@ putmail(dp, rp, fdmail, timestring, file)
 
 	/* append From_ line and print out the header */
 	eofindex = ftell(fp);
+
 	if (!mmdf_mode &&
 	    *(rp->addr->user) != TO_PIPE && eofindex > 0L &&
 	    fseek(fp, -2L, L_XTND) != -1) {
@@ -1009,6 +1047,7 @@ putmail(dp, rp, fdmail, timestring, file)
 	  eofindex = 0;
 	}
 
+
 	if (!(convert_qp && qp_to_8bit(rp)))
 	  convert_qp = 0; /* If malloc failed, clear this too,
 			     if it was not set, it doesn't change.. */
@@ -1033,13 +1072,19 @@ putmail(dp, rp, fdmail, timestring, file)
 	  fclose(fp);
 	  return NULL;
 	}
-	if (appendlet(dp, rp, fp, file, is_mime) == EOF) {
+	lastch = appendlet(dp, rp, fp, file, is_mime);
+	if (lastch == EOF) {
 	  /* XX: should I really do this? */
 	  fflush(fp);
 	  ftruncate(fileno(fp), (u_long)eofindex);
 	  fclose(fp);
 	  return NULL;
 	}
+	/* ---- knock-off unnecessary trailing newlines ????      ---
+	   ---- if lastch == '\n', adding ONE newline is enough ? ---
+	   if (lastch != '\n' && !ferror(fp))
+	   fputc('\n', fp);
+	 */
 	if (fputc('\n', fp) == EOF || fflush(fp) == EOF) {
 	  notaryreport(NULL,NULL,NULL);
 	  DIAGNOSTIC3(rp, EX_IOERR,
@@ -1164,9 +1209,9 @@ program(dp, rp, timestring)
 	  dup(in[1]);		/* in fd 2 */
 	  close(out[0]);
 	  close(in[1]);
-	  signal(SIGINT, SIG_IGN);
-	  signal(SIGHUP, SIG_IGN);
-	  signal(SIGTERM, SIG_DFL);
+	  SIGNAL_IGNORE(SIGINT);
+	  SIGNAL_IGNORE(SIGHUP);
+	  SIGNAL_HANDLE(SIGTERM, SIG_DFL);
 	  /*
 	   * Note that argv[0] is set to the command we are running.
 	   * That way, we should get some better error messages, at
@@ -1289,7 +1334,7 @@ creatembox(rp, filep, uid, gid, pw)
 #else
 	      struct utimbuf tv;
 	      stat(*filep,&st); /* This by all propability will not fail.. */
-	      tv.actime  = 1; /* Not zero, but close.. */
+	      tv.actime  = 1;	/* Not zero, but close.. */
 	      tv.modtime = st.st_mtime;
 	      utime(*filep, &tv);
 #endif
@@ -1299,8 +1344,11 @@ creatembox(rp, filep, uid, gid, pw)
 	}
 
 	/* assert *filep != NULL */
-	if (fd == -1)
+	if (fd == -1) {
+	  notaryreport(rp->addr->user, "failed",
+		       "466 (can't create user spool mailbox file)");
 	  DIAGNOSTIC(rp, EX_CANTCREAT, "can't create \"%s\"", *filep);
+	}
 	/* otherwise the message was printed by createfile() */
 	free(*filep);
 	return 0;
@@ -1321,22 +1369,29 @@ createfile(rp, file, uid)
 	  if ((cp = strrchr(file, '/')) != NULL) {
 	    *cp = '\0';
 	    if (exstat(rp, file, &st, stat) < 0) {
+	      *cp = '/';
+	      sprintf(msg,"466 (*INVOCATION BUG* Can't create user spool mailbox file: \"%s\", Directory stat() error: %s)",
+		      file,strerror(errno));
+	      *cp = 0;
+	      notaryreport(rp->addr->user, "failed", msg);
 	      DIAGNOSTIC(rp, i, "stat failed on %s", file);
 	      return -2;
 	    }
+	    *cp = '/';
 	    if (st.st_mode & 020) { /* group writable? */
 	      if (!setupuidgid(rp, uid, st.st_gid)) {
+		notaryreport(rp->addr->user, "failed",
+			     "466 (*INVOCATION BUG* The mailbox directory is group writable, but can't change my uid to it)");
 		DIAGNOSTIC(rp, i,
 			   "failed changing group id to create file in \"%s\"",
 			   file);
 		return -2;
 	      }
 	    }
-	    *cp = '/';
 	  }
 	  if ((fd = open(file, O_RDWR|O_CREAT|O_EXCL, MAILMODE)) < 0) {
-	    setrootuid(rp);
 	    saverrno = errno;
+	    setrootuid(rp);
 	  } else {
 	    setrootuid(rp);
 	    return fd;
@@ -1354,6 +1409,8 @@ createfile(rp, file, uid)
 	else /* if (saverrno == ENOENT) */
 	  i = EX_CANTCREAT;
 	/* convoluted to maintain 4 arguments to DIAGNOSTIC */
+	sprintf(msg, "466 (error [%s] creating \"%%s\")", strerror(saverrno));
+	notaryreport(rp->addr->user, "failed",msg);
 	sprintf(msg, "error (%s) creating \"%%s\"", strerror(saverrno));
 	DIAGNOSTIC(rp, i, msg, file);
 	return -2;
@@ -1440,6 +1497,9 @@ setrootuid(rp)
 
 /*
  * appendlet - append letter to file pointed at by fd
+ *
+ *	Return the last character written..
+ *
  */
 int
 appendlet(dp, rp, fp, file, ismime)
@@ -1449,6 +1509,7 @@ appendlet(dp, rp, fp, file, ismime)
 	int ismime;
 	char *file;
 {
+	int lastch = -1;
 #ifndef	USE_MMAP
 	static char buffer[BUFSIZ+1];
 	static char *bufend = buffer+BUFSIZ;
@@ -1459,9 +1520,9 @@ appendlet(dp, rp, fp, file, ismime)
 	int mfd = dp->msgfd;
 
 	if (ismime)
-	  writemimebuf((FILE *)NULL, (char *)NULL, 0);	/* initialize state */
+	  writemimeline(NULL, NULL, 0, NULL);	/* initialize state */
 	else
-	  writebuf((FILE *)NULL, (char *)NULL, 0);	/* initialize state */
+	  writebuf(NULL, NULL, 0, NULL);	/* initialize state */
 
 	if (ismime) {
 	  /* can we use cache of message body data ? */
@@ -1478,7 +1539,7 @@ appendlet(dp, rp, fp, file, ismime)
 	      if (*s == '\n' && readalready > 0) {
 		++s; ++linelen; --readalready;
 	      }
-	      if (writemimebuf(fp, s0, linelen) != linelen) {
+	      if (writemimeline(fp, s0, linelen, &lastch) != linelen) {
 		DIAGNOSTIC(rp, EX_IOERR,
 			   "write to \"%s\" failed", file);
 		return EOF;
@@ -1486,7 +1547,7 @@ appendlet(dp, rp, fp, file, ismime)
 	      s0 = s;
 	    }
 	    rp->status = EX_OK;
-	    return 0;
+	    return lastch;
 	  }
 
 	  mfp = fdopen(mfd,"r");
@@ -1505,7 +1566,7 @@ appendlet(dp, rp, fp, file, ismime)
 	    if (i == BUFSIZ && buffer[BUFSIZ-1] != '\n')
 	      ismime = 0;
 	    /* Ok, write the line */
-	    if (writemimebuf(fp, buffer, i) != i) {
+	    if (writemimeline(fp, buffer, i, &lastch) != i) {
 	      DIAGNOSTIC(rp, EX_IOERR,
 			 "write to \"%s\" failed", file);
  	      MFPCLOSE;
@@ -1525,13 +1586,13 @@ appendlet(dp, rp, fp, file, ismime)
 
 	  /* can we use cache of message body data */
 	  if (readalready != 0) {
-	    if (writebuf(fp, buffer, readalready) != readalready) {
+	    if (writebuf(fp, buffer, readalready, &lastch) != readalready) {
 	      DIAGNOSTIC(rp, EX_IOERR,
 			 "write to \"%s\" failed", file);
 	      return EOF;
 	    }
 	    rp->status = EX_OK;
-	    return 0;
+	    return lastch;
 	  }
 
 	  /* we are assumed to be positioned properly
@@ -1543,7 +1604,7 @@ appendlet(dp, rp, fp, file, ismime)
 			 "read error from message file", 0);
 	      return EOF;
 	    }
-	    if (writebuf(fp, buffer, i) != i) {
+	    if (writebuf(fp, buffer, i, &lastch) != i) {
 	      DIAGNOSTIC(rp, EX_IOERR,
 			 "write to \"%s\" failed", file);
 	      return EOF;
@@ -1560,13 +1621,13 @@ appendlet(dp, rp, fp, file, ismime)
 	s = dp->let_buffer + dp->msgbodyoffset;
 	if (ismime) {
 	  int i;
-	  writemimebuf((FILE *)NULL, (char *)NULL, 0);	/* initialize state */
+	  writemimeline(NULL, NULL, 0, NULL);	/* initialize state */
 	  while (s < dp->let_end) {
 	    char *s2 = s;
 	    i = 0;
 	    while (s2 < dp->let_end && *s2 != '\n') ++s2, ++i;
 	    if (s2 < dp->let_end && *s2 == '\n') ++s2, ++i;
-	    if (writemimebuf(fp, s, i) != i) {
+	    if (writemimeline(fp, s, i, &lastch) != i) {
 	      DIAGNOSTIC(rp, EX_IOERR,
 			 "write to \"%s\" failed", file);
 	      return EOF;
@@ -1574,15 +1635,15 @@ appendlet(dp, rp, fp, file, ismime)
 	    s = s2;
 	  }
 	} else {
-	  writebuf((FILE *)NULL, (char *)NULL, 0);	/* initialize state */
-	  if (writebuf(fp, s, dp->let_end - s) != (dp->let_end - s)) {
+	  writebuf(NULL, NULL, 0, NULL);	/* initialize state */
+	  if (writebuf(fp, s, dp->let_end - s, &lastch) != (dp->let_end - s)) {
 	      DIAGNOSTIC(rp, EX_IOERR,
 			 "write to \"%s\" failed", file);
 	      return EOF;
 	  }
 	}
 #endif
-	return 0;
+	return lastch;
 }
 
 
@@ -1741,10 +1802,11 @@ rbiff(nbp)
  */
 
 int
-writebuf(fp, buf, len)
+writebuf(fp, buf, len, lastchp)
 	FILE *fp;
 	char *buf;
 	int len;
+	int *lastchp;
 {
 	register char *cp, *s;
 	register int n;
@@ -1783,6 +1845,8 @@ writebuf(fp, buf, len)
 
 	expect = save;
 
+	*lastchp = buf[len-1];
+
 	for (cp = buf, n = len, tlen = 0, s = NULL; n > 0; --n, ++cp) {
 	  register int c = *cp;
 	  ++tlen;
@@ -1833,10 +1897,11 @@ writebuf(fp, buf, len)
 }
 
 int
-writemimebuf(fp, buf, len)
+writemimeline(fp, buf, len, lastchp)
 	FILE *fp;
 	char *buf;
 	int len;
+	int *lastchp;
 {
 	register char *cp, *s;
 	register int n;
@@ -1872,14 +1937,14 @@ writemimebuf(fp, buf, len)
 
 	/* ------------------------------------------------------------ */
 	/* MIME format processing, the input is LINE BY LINE */
-	if (mime_boundary != NULL && n > mime_boundary_len+2 &&
+	if (mime_boundary != NULL && len > mime_boundary_len+2 &&
 	    buf[0] == '-' && buf[1] == '-' &&
 	    strncmp(buf+2,mime_boundary,mime_boundary_len) == 0) {
 	  /* ------------------------------------------------------------ */
 	  /* XX: process MIME boundary! */
-	  i = fwrite(buf, 1, len, fp);
-	  if (i != len) return -1;
-	  return len;
+	  *lastchp = buf[len-1];
+	  return fwrite(buf, 1, len, fp);
+
 	} else if (convert_qp) {
 	  /* ------------------------------------------------------------ */
 	  /* Process Q-P chars of this line */
@@ -1921,17 +1986,21 @@ writemimebuf(fp, buf, len)
 	  /* Uhh... Stored the conversion result into buf2[], length  tlen */
 	  if (!mmdf_mode && buf2[0] == 'F' && strncmp(buf2,"From ",5)==0)
 	    putc('>',fp);
-	  i = fwrite(buf2, 1, tlen, fp);
+	  if (tlen > 0) {
+	    i = fwrite(buf2, 1, tlen, fp);
+	    *lastchp = buf2[tlen-1];
+	  } else
+	    i = 0;
 	  if (i != tlen) return -1;
-	  return len;		/* Return the incoming length, NOT the true length! */
+	  return len;		/* Return the incoming length,
+				   NOT the true length! */
 	}
 	/* ------------------------------------------------------------ */
 	/* Well, no other processings known.. */
 	if (!mmdf_mode && buf[0] == 'F' && strncmp(buf,"From ",5)==0)
 	  putc('>',fp);
-	i = fwrite(buf, 1, len, fp);
-	if (i != len) return -1;
-	return len;
+	*lastchp = buf[len-1];
+	return fwrite(buf, 1, len, fp);
 }
 
 
