@@ -8,13 +8,13 @@
 #include <stdio.h>
 #include <ctype.h>
 #include <pwd.h>
-#include <signal.h>
+#include "zmsignal.h"
 #include <sysexits.h>
 #include <varargs.h>
 #include <sys/file.h>
 #include <sys/param.h>
 #include <sys/stat.h>
-#include <mail.h>
+#include "mail.h"
 #include "malloc.h"
 #include "ta.h"
 #include NDIR_H
@@ -46,6 +46,7 @@ int	maxwidth = 0;
 
 int	D_alloc = 0;		/* Memory debugging */
 
+extern SIGNAL_TYPE wantout();
 #ifndef MALLOC_TRACE
 extern univptr_t emalloc();
 extern univptr_t erealloc();
@@ -126,14 +127,19 @@ main(argc, argv)
 	struct ctldesc *dp;
 	int errflg, c;
 	struct maildesc *mp;
+	SIGNAL_TYPE (*oldsig)();
 
-	if (signal(SIGINT, SIG_IGN) != SIG_IGN)
-	  signal(SIGINT, wantout);
-	if (signal(SIGHUP, SIG_IGN) != SIG_IGN)
-	  signal(SIGHUP, wantout);
-	if (signal(SIGTERM, SIG_IGN) != SIG_IGN)
-	  signal(SIGTERM, wantout);
-	signal(SIGPIPE, SIG_IGN);
+	SIGNAL_HANDLESAVE(SIGINT, SIG_IGN, oldsig);
+	if (oldsig != SIG_IGN)
+	  SIGNAL_HANDLE(SIGINT, wantout);
+	SIGNAL_HANDLESAVE(SIGTERM, SIG_IGN, oldsig);
+	if (oldsig != SIG_IGN)
+	  SIGNAL_HANDLE(SIGTERM, wantout);
+	SIGNAL_HANDLESAVE(SIGHUP, SIG_IGN, oldsig);
+	if (oldsig != SIG_IGN)
+	  SIGNAL_HANDLE(SIGHUP, wantout);
+
+	SIGNAL_IGNORE(SIGPIPE);
 
 	if ((progname = strrchr(argv[0], '/')) == NULL)
 	  progname = argv[0];
@@ -567,7 +573,7 @@ appendlet(dp, mp, fp, verboselog, convertmode)
 
 	register int i;
 	register int bufferfull;
-	int lastwasnl;
+	int lastch;
 #ifndef USE_MMAP
 	int mfd = dp->msgfd;
 #endif
@@ -577,9 +583,10 @@ appendlet(dp, mp, fp, verboselog, convertmode)
 #ifndef USE_MMAP
 	/* can we use cache of message body data */
 	if (convertmode == _CONVERT_NONE && readalready != 0) {
+	  lastch = let_buffer[readalready-1];
 	  if (writebuf(mp, fp, let_buffer, readalready) != readalready)
 	    return EX_IOERR;
-	  if (let_buffer[readalready-1] != '\n')
+	  if (lastch != '\n')
 	    if (writebuf(mp, fp, "\n", 1) != 1)
 	      return EX_IOERR;
 	  return EX_OK;
@@ -587,7 +594,8 @@ appendlet(dp, mp, fp, verboselog, convertmode)
 #endif
 
 	/* we are assumed to be positioned properly at start of message body */
-	bufferfull = lastwasnl = 0;
+	bufferfull = 0;
+	lastch = -1;
 	if (convertmode == _CONVERT_NONE) {
 #ifndef	USE_MMAP
 	  while ((i = read(mfd, let_buffer, sizeof let_buffer)) != 0) {
@@ -597,7 +605,7 @@ appendlet(dp, mp, fp, verboselog, convertmode)
 #endif
 	    if (i < 0)
 	      return EX_IOERR;
-	    lastwasnl = (let_buffer[i-1] == '\n');
+	    lastch = let_buffer[i-1];
 	    if (writebuf(mp, fp, let_buffer, i) != i)
 	      return EX_IOERR;
 	    readalready = i;
@@ -614,7 +622,7 @@ appendlet(dp, mp, fp, verboselog, convertmode)
 	  char iobuf[BUFSIZ];
 	  FILE *mfp = fdopen(mfd,"r");
 
-	  writemimeline(mp, fp, (char *)NULL, 0, 0);
+	  writemimeline(mp, fp, (char *)NULL, 0, 0, NULL);
 	  setvbuf(mfp, iobuf, _IOFBF, sizeof(iobuf));
 
 #define MFPCLOSE i = dup(mfd); fclose(mfp); dup2(i,mfd); close(i);
@@ -628,27 +636,27 @@ appendlet(dp, mp, fp, verboselog, convertmode)
 	  /* we are assuming to be positioned properly
 	     at the start of the message body */
 	  bufferfull = 0;
-	  lastwasnl = 0;
+	  lastch = -1;
 	  i = 0;
 	  for (;;) {
 #ifndef USE_MMAP
 	    if ((i = cfgets(let_buffer, sizeof(let_buffer), mfp)) == EOF)
 	      break;
-	    lastwasnl = (let_buffer[i-1] == '\n');
 #else
 	    char *let_buffer = s, *s2 = s;
 	    if (s >= dp->let_end) break;	/* "EOF" */
 	    i = 0;
 	    while (s2 < dp->let_end && *s2 != '\n')
 	      ++s2, ++i;
-	    if ((lastwasnl = (*s2 == '\n')))
+	    if ((lastch = *s2) == '\n')
 	      ++s2, ++i;
 	    s = s2;
 #endif
 	    /* It MAY be malformed -- if it has a BUFSIZ length
 	       line in it, IT CAN'T BE MIME  :-/		*/
 	    /* Ok, write the line */
-	    if (writemimeline(mp, fp, let_buffer, i, convertmode) != i) {
+	    if (writemimeline(mp, fp, let_buffer, i,
+			      convertmode, &lastch) != i) {
 	      return EX_IOERR;
 	    }
 	  }
@@ -662,7 +670,7 @@ appendlet(dp, mp, fp, verboselog, convertmode)
 	}
 
 	/* we must make sure the last thing we transmit is a CRLF sequence */
-	if (!lastwasnl)
+	if (lastch != '\n')
 	  writebuf(mp, fp, "\n", 1);
 
 #ifndef USE_MMAP
@@ -773,11 +781,11 @@ writebuf(mp, fp, buf, len)
 }
 
 int
-writemimeline(mp, fp, buf, len, convertmode)
+writemimeline(mp, fp, buf, len, convertmode, lastchp)
 	struct maildesc *mp;
 	FILE *fp;
 	char *buf;
-	int len, convertmode;
+	int len, convertmode, *lastchp;
 {
 	/* `convertmode' controls the behaviour of the message conversion:
 	     _CONVERT_NONE (0): send as is
