@@ -13,7 +13,7 @@
 #include <ctype.h>
 #include <errno.h>
 #include <pwd.h>
-#include <signal.h>
+#include "zmsignal.h"
 #include <sysexits.h>
 /* #include <strings.h> */ /* poorly portable.. */
 #include <varargs.h>
@@ -26,7 +26,7 @@
 #include <netdb.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
-#include <mail.h>
+#include "mail.h"
 #include "ta.h"
 #include "malloc.h"
 
@@ -129,19 +129,6 @@ typedef char msgdata;
 #define	PROGNAME	"smtpclient"	/* for logging */
 #define	CHANNEL		"smtp"	/* the default channel name we deliver for */
 
-/* SV_INTERRUPT is defined in <sys/signal.h> on SunOS */
-#ifdef	SV_INTERRUPT
-#define	SIGNAL_SYSCALLINTERRUPT(X,Y)	\
-	{	struct sigvec sv; \
-		sigvec(X, (struct sigvec *)NULL, &sv); \
-		sv.sv_handler = Y; \
-		sv.sv_flags |= SV_INTERRUPT; \
-		sigvec(X, &sv, (struct sigvec *)NULL); \
-	}
-#else	/* !SV_INTERRUPT */
-#define	SIGNAL_SYSCALLINTERRUPT(X,Y)	signal(X,Y)
-#endif	/* SV_INTERRUPT */
-
 #ifndef	MAXHOSTNAMELEN
 #define	MAXHOSTNAMELEN 64
 #endif	/* MAXHOSTNAMELEN */
@@ -217,7 +204,7 @@ extern char *optarg;
 extern int optind;
 extern int deliver __((struct ctldesc *dp, FILE *smtpfp[2], FILE *verboselog, struct rcpt *startrp, struct rcpt *endrp));
 extern int writebuf __((FILE *fp, char *buf, int len));
-extern int writemimeline __((FILE *fp, char *buf, int len, int convertmode));
+extern int writemimeline __((FILE *fp, char *buf, int len, int convertmode, int*lastchp));
 extern int appendlet __((struct ctldesc *dp, FILE *fp, FILE *verboselog, int hsize, int msize, int convertmode));
 extern int smtpopen __((char *host, FILE *fparr[], FILE *verboselog));
 extern int smtpconn __((char *host, FILE *fparr[], FILE *verboselog));
@@ -274,6 +261,7 @@ main(argc, argv)
 #ifdef	BIND
 	int checkmx = 0;	/* check all destination hosts for MXness */
 #endif	/* BIND */
+	SIGNAL_TYPE (*oldsig)();
 
 	smtpfp[0] = NULL;
 	smtpfp[1] = NULL;
@@ -286,18 +274,23 @@ main(argc, argv)
 	eocmdline = cmdline;
 	for (i = 0; i < argc; ++i)
 	  eocmdline += strlen(argv[i]) + 1;
-	if (signal(SIGINT, SIG_IGN) != SIG_IGN)
-	  signal(SIGINT, wantout);
-	if (signal(SIGHUP, SIG_IGN) != SIG_IGN)
-	  signal(SIGHUP, wantout);
-	if (signal(SIGTERM, SIG_IGN) != SIG_IGN)
-	  signal(SIGTERM, wantout);
+
+	SIGNAL_HANDLESAVE(SIGINT, SIG_IGN, oldsig);
+	if (oldsig != SIG_IGN)
+	  SIGNAL_HANDLE(SIGINT, wantout);
+	SIGNAL_HANDLESAVE(SIGTERM, SIG_IGN, oldsig);
+	if (oldsig != SIG_IGN)
+	  SIGNAL_HANDLE(SIGTERM, wantout);
+	SIGNAL_HANDLESAVE(SIGHUP, SIG_IGN, oldsig);
+	if (oldsig != SIG_IGN)
+	  SIGNAL_HANDLE(SIGHUP, wantout);
+
 #if 0
-	SIGNAL_SYSCALLINTERRUPT(SIGPIPE, sig_pipe);
+	SIGNAL_HANDLE(SIGPIPE, sig_pipe);
 #else
-	signal(SIGPIPE, SIG_IGN);
+	SIGNAL_IGNORE(SIGPIPE);
 #endif
-	SIGNAL_SYSCALLINTERRUPT(SIGALRM, sig_alarm);
+	SIGNAL_HANDLE(SIGALRM, sig_alarm);
 	timeout = TIMEOUT;
 
 	progname = PROGNAME;
@@ -785,6 +778,8 @@ appendlet(dp, fp, verboselog, hsize, msize, convertmode)
 	register int bufferfull;
 	int position = dp->msgbodyoffset;
 	int lastwasnl;
+	int lastch = '\n'; /* WriteMIMELine() can decode-QP and then the
+			      "lastwasnl" is no longer valid .. */
 
 	writebuf(fp, (char *)NULL, 0);	/* magic init. */
 
@@ -866,7 +861,8 @@ appendlet(dp, fp, verboselog, hsize, msize, convertmode)
 
 #define MFPCLOSE
 #endif
-	  writemimeline(fp, (char *)NULL, 0, 0);
+
+	  writemimeline(fp, (char *)NULL, 0, 0, NULL);
 
 	  /* we are assuming to be positioned properly
 	     at the start of the message body */
@@ -893,8 +889,8 @@ appendlet(dp, fp, verboselog, hsize, msize, convertmode)
 #endif
 	    /* XX: Detect multiparts !! */
 
-	    /* Ok, write the line */
-	    if (writemimeline(fp, let_buffer, i, convertmode) != i) {
+	    /* Ok, write the line -- decoding QP can alter the "lastwasnl" */
+	    if (writemimeline(fp, let_buffer, i, convertmode, &lastch) != i) {
 	      sprintf(remotemsg,
 		      "500 (let_buffer write IO-error! [%s] DATA %d/%d [%d%%])",
 		      strerror(errno), hsize, msize, (hsize*100+msize/2)/msize);
@@ -916,7 +912,7 @@ appendlet(dp, fp, verboselog, hsize, msize, convertmode)
 	} /* ... end of conversion modes */
 
 	/* we must make sure the last thing we transmit is a CRLF sequence */
-	if (!lastwasnl)
+	if (!lastwasnl || lastch != '\n')
 	  if (writebuf(fp, "\n", 1) != 1) {
 	    if (gotalarm) {
 	      sprintf(remotemsg,"500 (let_buffer write timeout!  DATA %d/%d [%d%%])",
@@ -1007,10 +1003,10 @@ writebuf(fp, buf, len)
 
 
 int
-writemimeline(fp, buf, len, convertmode)
+writemimeline(fp, buf, len, convertmode, lastcharp)
 	FILE *fp;
 	char *buf;
-	int len;
+	int len, *lastcharp;
 	int convertmode;
 {
 	register char *cp;
@@ -1040,11 +1036,12 @@ writemimeline(fp, buf, len, convertmode)
 	  /* No magics here.. we are linemode.. */
 	  return 0;
 	}
+	*lastcharp = -1;
 	for (cp = buf, n = len; n > 0; --n, ++cp) {
 	  int c = *(unsigned char *)cp;
 	  ++column;
 	  if (convertmode == _CONVERT_8BIT) {
-	    if (qp_chrs == 0 && c == '=') {
+	    if (c == '=' && qp_chrs == 0) {
 	      qp_val = 0;
 	      qp_chrs = 2;
 	      continue;
@@ -1073,11 +1070,13 @@ writemimeline(fp, buf, len, convertmode)
 	      else
 		continue;
 	    }
+	    *lastcharp = c;
 	  } else {
 	    if (qp_conv && column > 70 && c != '\n') {
 	      putc('=',fp);
 	      putc('\r',fp);
 	      putc('\n',fp);
+	      *lastcharp = '\n';
 	      /* if (verboselog) fprintf(verboselog,"=\n"); */
 	      column = -1;
 	    }
@@ -1087,6 +1086,7 @@ writemimeline(fp, buf, len, convertmode)
 	      putc('=',fp);
 	      putc(i2h[(c >> 4) & 15],fp);
 	      putc(i2h[(c)      & 15],fp);
+	      *lastcharp = i2h[(c) & 15];
 	      column += 3;
 	      /* if (verboselog) fprintf(verboselog,"=%02X",c); */
 	      continue;
@@ -1116,6 +1116,7 @@ writemimeline(fp, buf, len, convertmode)
 	    strcpy(remotemsg, "write error 4");
 	    return EOF;
 	  }
+	  *lastcharp = c;
 	  /* if (verboselog) putc(c, verboselog); */
 
 	}
